# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2025, by Samuel Williams.

require_relative "body/chunked"
require_relative "body/fixed"
require_relative "body/remainder"

module Protocol
	module HTTP1
		# A collection of classes for handling HTTP/1.1 request and response bodies.
		module Body
		end
	end
end
