# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2019-2025, by Samuel Williams.

require_relative "http1/version"
require_relative "http1/connection"

# @namespace
module Protocol
	# @namespace
	module HTTP1
	end
end
